package org.dimdev.dimdoors.block.door.data;

import java.util.*;
import java.util.function.Consumer;

import com.google.common.collect.ImmutableMap;
import com.google.gson.JsonElement;
import com.google.gson.JsonObject;
import com.google.gson.JsonPrimitive;
import org.dimdev.dimdoors.block.door.DimensionalDoorBlock;
import org.dimdev.dimdoors.block.entity.EntranceRiftBlockEntity;
import org.dimdev.dimdoors.item.DimensionalDoorItem;
import org.dimdev.dimdoors.item.ItemExtensions;
import net.fabricmc.fabric.api.object.builder.v1.block.FabricBlockSettings;
import net.fabricmc.fabric.api.util.TriState;
import net.minecraft.class_156;
import net.minecraft.class_1761;
import net.minecraft.class_1792;
import net.minecraft.class_1814;
import net.minecraft.class_2248;
import net.minecraft.class_2378;
import net.minecraft.class_2960;
import org.dimdev.dimdoors.item.ModItems;
import org.dimdev.dimdoors.mixin.accessor.ItemGroupAccessor;

@SuppressWarnings("OptionalUsedAsFieldOrParameterType")
public final class DoorData implements AutoCloseable {
	private static final Map<String, class_1761> itemGroupCache = new HashMap<>();
	public static final Set<class_2248> PARENT_BLOCKS = new HashSet<>();
	public static final Set<class_1792> PARENT_ITEMS = new HashSet<>();
	public static final List<class_2248> DOORS = new ArrayList<>();
	private final String id;
	private final UnbakedItemSettings itemSettings;
	private final Optional<String> itemGroup;
	private final UnbakedBlockSettings blockSettings;
	private final RiftDataList riftDataList;
	private final boolean hasToolTip;
	private boolean closed = false;

	public static DoorData fromJson(JsonObject json) {
		try {
			String id = json.get("id").getAsString();
			UnbakedItemSettings itemSettings = UnbakedItemSettings.fromJson(json.getAsJsonObject("itemSettings"));
			Optional<String> itemGroup = Optional.ofNullable(json.getAsJsonPrimitive("itemGroup")).map(JsonPrimitive::getAsString);
			UnbakedBlockSettings blockSettings = UnbakedBlockSettings.fromJson(json.getAsJsonObject("blockSettings"));
			RiftDataList riftDataList = RiftDataList.fromJson(json.getAsJsonArray("riftData"));
			boolean hasToolTip = json.has("hasToolTip") && json.getAsJsonPrimitive("hasToolTip").getAsBoolean();
			return new DoorData(id, itemSettings, itemGroup, blockSettings, riftDataList, hasToolTip);
		} catch (RuntimeException e) {
			throw new RuntimeException("Caught exception while deserializing " + json.toString(), e);
		}
	}

	public DoorData(String id, UnbakedItemSettings itemSettings, UnbakedBlockSettings blockSettings, RiftDataList riftDataList) {
		this(id, itemSettings, Optional.empty(), blockSettings, riftDataList, false);
	}

	public DoorData(String id, UnbakedItemSettings itemSettings, UnbakedBlockSettings blockSettings, RiftDataList riftDataList, boolean hasToolTip) {
		this(id, itemSettings, Optional.empty(), blockSettings, riftDataList, hasToolTip);
	}

	public DoorData(String id, UnbakedItemSettings itemSettings, Optional<String> itemGroup, UnbakedBlockSettings blockSettings, RiftDataList riftDataList, boolean hasToolTip) {
		this.id = id;
		this.itemSettings = itemSettings;
		this.itemGroup = itemGroup;
		this.blockSettings = blockSettings;
		this.riftDataList = riftDataList;
		this.hasToolTip = hasToolTip;
	}

	public JsonObject toJson(JsonObject json) {
		json.addProperty("id", this.id);
		json.add("itemSettings", this.itemSettings.toJson(new JsonObject()));
		itemGroup.ifPresent(s -> json.add("itemGroup", new JsonPrimitive(s)));
		json.add("blockSettings", this.blockSettings.toJson(new JsonObject()));
		json.add("riftData", this.riftDataList.toJson());
		json.addProperty("hasToolTip", this.hasToolTip);
		return json;
	}

	private Consumer<? super EntranceRiftBlockEntity> createSetupFunction() {
		RiftDataList riftDataList = this.riftDataList;

		return rift -> {
			RiftDataList.OptRiftData riftData = riftDataList.getRiftData(rift);
			riftData.getDestination().ifPresent(rift::setDestination);
			riftData.getProperties().ifPresent(rift::setProperties);
		};
	}

	public String getId() {
		return id;
	}

	public UnbakedItemSettings getItemSettings() {
		return itemSettings;
	}

	public UnbakedBlockSettings getBlockSettings() {
		return blockSettings;
	}

	public RiftDataList getRiftDataList() {
		return riftDataList;
	}

	public boolean hasToolTip() {
		return hasToolTip;
	}

	@Override
	public void close() {
		if (closed) {
			throw new UnsupportedOperationException("Already Closed");
		}

		class_1792 parentItem = class_2378.field_11142.method_10223(new class_2960(this.itemSettings.parent));
		PARENT_ITEMS.add(parentItem);
		class_1792.class_1793 itemSettings = ItemExtensions.getSettings(parentItem);
		this.itemSettings.maxCount.ifPresent(itemSettings::method_7889);
		this.itemSettings.maxDamage.ifPresent(itemSettings::method_7898);
		this.itemSettings.rarity.ifPresent(itemSettings::method_7894);
		this.itemSettings.fireproof.map(b -> {
			if (!b) return false;
			itemSettings.method_24359();
			return false;
		});
		class_1761 group = null;
		if (itemGroup.isPresent()) {
			String groupString = itemGroup.get();
			if (itemGroupCache.containsKey(groupString)) {
				group = itemGroupCache.get(groupString);
			} else {
				for (class_1761 g : class_1761.field_7921) {
					if (((ItemGroupAccessor) g).getId().equals(groupString)) {
						group = g;
						itemGroupCache.put(groupString, group);
						break;
					}
				}
			}
		}
		itemSettings.method_7892(group != null ? group : ModItems.DIMENSIONAL_DOORS);

		class_2248 parentBlock = class_2378.field_11146.method_10223(new class_2960(this.blockSettings.parent));
		PARENT_BLOCKS.add(parentBlock);
		FabricBlockSettings blockSettings = FabricBlockSettings.copyOf(parentBlock);
		this.blockSettings.luminance.ifPresent(blockSettings::luminance);
		class_2960 id = new class_2960(this.id);
		class_2248 doorBlock = new DimensionalDoorBlock(blockSettings);
		class_1792 doorItem = new DimensionalDoorItem(doorBlock, itemSettings, createSetupFunction(), hasToolTip);
		class_2378.method_10230(class_2378.field_11146, id, doorBlock);
		class_2378.method_10230(class_2378.field_11142, id, doorItem);
		DOORS.add(doorBlock);
		class_1792.field_8003.put(doorBlock, doorItem);
		this.closed = true;
	}

	@SuppressWarnings("OptionalUsedAsFieldOrParameterType")
	public static final class UnbakedItemSettings {
		private static final Map<String, class_1814> RARITIES = class_156.method_654(ImmutableMap.<String, class_1814>builder(), b -> {
			for (class_1814 rarity : class_1814.values()) {
				b.put(rarity.name().toLowerCase(), rarity);
			}
		}).build();
		private final String parent;
		private final OptionalInt maxCount;
		private final OptionalInt maxDamage;
		private final Optional<class_1814> rarity;
		private final TriState fireproof;

		public static UnbakedItemSettings fromJson(JsonObject json) {
			String parent = Optional.ofNullable(json.get("parent")).map(JsonElement::getAsString).orElseThrow(() -> new RuntimeException("Missing parent key in " + json.toString()));
			OptionalInt maxCount = Optional.ofNullable(json.get("maxCount")).map(JsonElement::getAsInt).map(OptionalInt::of).orElse(OptionalInt.empty());
			OptionalInt maxDamage = Optional.ofNullable(json.get("maxDamage")).map(JsonElement::getAsInt).map(OptionalInt::of).orElse(OptionalInt.empty());
			Optional<class_1814> rarity = Optional.ofNullable(json.get("rarity")).map(JsonElement::getAsString).map(String::toLowerCase).map(RARITIES::get).map(Objects::requireNonNull);
			TriState fireproof = Optional.ofNullable(json.get("fireproof")).map(JsonElement::getAsBoolean).map(TriState::of).orElse(TriState.DEFAULT);
			return new UnbakedItemSettings(parent, maxCount, maxDamage, rarity, fireproof);
		}

		public UnbakedItemSettings(String parent, OptionalInt maxCount, OptionalInt maxDamage, Optional<class_1814> rarity, TriState fireproof) {
			this.parent = parent;
			this.maxCount = maxCount;
			this.maxDamage = maxDamage;
			this.rarity = rarity;
			this.fireproof = fireproof;
		}

		public JsonObject toJson(JsonObject json) {
			json.addProperty("parent", parent);
			maxCount.ifPresent(s -> json.addProperty("maxCount", s));
			maxDamage.ifPresent(s -> json.addProperty("maxDamage", s));
			rarity.ifPresent(s -> json.addProperty("rarity", s.name().toLowerCase()));
			fireproof.map(b -> {
				json.addProperty("fireproof", b);
				return b;
			});
			return json;
		}
	}

	@SuppressWarnings("OptionalUsedAsFieldOrParameterType")
	public static final class UnbakedBlockSettings {
		private final String parent;
		private final OptionalInt luminance;

		public static UnbakedBlockSettings fromJson(JsonObject json) {
			String parent = Optional.ofNullable(json.get("parent")).map(JsonElement::getAsString).orElseThrow(() -> new RuntimeException("Missing parent block"));
			OptionalInt luminance = Optional.ofNullable(json.get("luminance")).map(JsonElement::getAsInt).map(OptionalInt::of).orElse(OptionalInt.empty());
			return new UnbakedBlockSettings(parent, luminance);
		}

		public UnbakedBlockSettings(String parent, OptionalInt luminance) {
			this.parent = parent;
			this.luminance = luminance;
		}

		public JsonObject toJson(JsonObject json) {
			json.addProperty("parent", parent);
			luminance.ifPresent(s -> json.addProperty("luminance", s));
			return json;
		}
	}
}
