package org.dimdev.dimdoors.pockets.virtual.reference;

import java.util.*;
import java.util.stream.Collectors;
import java.util.stream.StreamSupport;

import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import net.fabricmc.fabric.api.util.NbtType;
import net.minecraft.class_2338;
import net.minecraft.class_2487;
import net.minecraft.class_2499;
import net.minecraft.class_2520;
import net.minecraft.class_2791;
import net.minecraft.class_3193;
import net.minecraft.class_3300;
import net.minecraft.class_3738;
import net.minecraft.server.MinecraftServer;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.dimdev.dimdoors.DimensionalDoorsInitializer;
import org.dimdev.dimdoors.pockets.generator.LazyPocketGenerator;
import org.dimdev.dimdoors.pockets.generator.PocketGenerator;
import org.dimdev.dimdoors.pockets.modifier.LazyCompatibleModifier;
import org.dimdev.dimdoors.pockets.modifier.LazyModifier;
import org.dimdev.dimdoors.pockets.modifier.Modifier;
import org.dimdev.dimdoors.pockets.modifier.RiftManager;
import org.dimdev.dimdoors.pockets.virtual.AbstractVirtualPocket;
import org.dimdev.dimdoors.pockets.virtual.ImplementedVirtualPocket;
import org.dimdev.dimdoors.pockets.PocketGenerationContext;
import org.dimdev.dimdoors.api.util.math.Equation;
import org.dimdev.dimdoors.api.util.math.Equation.EquationParseException;
import org.dimdev.dimdoors.world.pocket.type.LazyGenerationPocket;
import org.dimdev.dimdoors.world.pocket.type.Pocket;

public abstract class PocketGeneratorReference extends AbstractVirtualPocket {
	private static final Logger LOGGER = LogManager.getLogger();

	protected String weight;
	protected Equation weightEquation;
	protected Boolean setupLoot;
	protected final List<Modifier> modifierList = Lists.newArrayList();
	protected final List<class_2487> addons = new ArrayList<>();

	private void parseWeight() {
		try {
			this.weightEquation = Equation.parse(weight);
		} catch (EquationParseException e) {
			LOGGER.debug("Defaulting to default weight equation for {}", this);
			LOGGER.debug("Exception Stacktrace", e);
			try {
				// FIXME: do we actually want to have it serialize to the broken String equation we input?
				this.weightEquation = Equation.newEquation(Equation.parse(DimensionalDoorsInitializer.getConfig().getPocketsConfig().defaultWeightEquation)::apply, stringBuilder -> stringBuilder.append(weight));
			} catch (EquationParseException equationParseException) {
				LOGGER.debug("Defaulting to default weight equation for {}", this);
				LOGGER.debug("Exception Stacktrace", e);
				// FIXME: do we actually want to have it serialize to the broken String equation we input?
				this.weightEquation = Equation.newEquation(stringDoubleMap -> (double) DimensionalDoorsInitializer.getConfig().getPocketsConfig().fallbackWeight, stringBuilder -> stringBuilder.append(weight));
			}
		}
	}

	@Override
	public ImplementedVirtualPocket fromNbt(class_2487 nbt, class_3300 manager) {
		if (nbt.method_10545("weight")) { // override referenced pockets weight
			this.weight = nbt.method_10558("weight");
			parseWeight();
		}

		if (nbt.method_10545("setup_loot")) setupLoot = nbt.method_10577("setup_loot");

		if (nbt.method_10545("modifiers")) {
			class_2499 modifiersNbt = nbt.method_10554("modifiers", 10);
			for (int i = 0; i < modifiersNbt.size(); i++) {
				modifierList.add(Modifier.deserialize(modifiersNbt.method_10602(i), manager));
			}
		}
		if (nbt.method_10545("modifier_references")) {
			class_2499 modifiersNbt = nbt.method_10554("modifier_references", NbtType.STRING);
			for (class_2520 nbtElement : modifiersNbt) {
				modifierList.add(Modifier.deserialize(nbtElement, manager));
			}
		}

		if (nbt.method_10573("addons", NbtType.LIST)) {
			class_2499 addonsNbt = nbt.method_10554("addons", 10);
			for (int i = 0; i < addonsNbt.size(); i++) {
				// TODO: something with the ResourceManager??? Probably need AddonBuilder now.
				addons.add(addonsNbt.method_10602(i));
			}
		}

		return this;
	}

	@Override
	protected class_2487 toNbtInternal(class_2487 nbt, boolean allowReference) {

		if (weight != null) nbt.method_10582("weight", weight);

		if (setupLoot != null) nbt.method_10556("setup_loot", setupLoot);

		class_2499 modifiersNbt = new class_2499();
		// TODO: deserialize with ResourceManager
		for (Modifier modifier : modifierList) {
			modifiersNbt.add(modifier.toNbt(new class_2487()));
		}
		if (modifiersNbt.size() > 0) nbt.method_10566("modifiers", modifiersNbt);

		class_2499 addonsNbt = new class_2499();
		// TODO: something with the ResourceManager??? Probably need AddonBuilder now.
		addonsNbt.addAll(addons);
		if (addonsNbt.size() > 0) nbt.method_10566("addons", addonsNbt);

		return nbt;
	}

	@Override
	public double getWeight(PocketGenerationContext parameters) {
		try {
			return weightEquation != null ? this.weightEquation.apply(parameters.toVariableMap(Maps.newHashMap())) : peekReferencedPocketGenerator(parameters).getWeight(parameters);
		} catch (RuntimeException e) {
			LOGGER.error(this.toString());
			throw new AssertionError(e);
		}
	}

	public void applyModifiers(PocketGenerationContext parameters, RiftManager manager) {
		for (Modifier modifier : modifierList) {
			modifier.apply(parameters, manager);
		}
	}

	public void applyModifiers(PocketGenerationContext parameters, Pocket.PocketBuilder<?, ?> builder) {
		for (Modifier modifier : modifierList) {
			modifier.apply(parameters, builder);
		}
	}

	@Override
	public Pocket prepareAndPlacePocket(PocketGenerationContext parameters) {
		PocketGenerator generator = getReferencedPocketGenerator(parameters);


		Pocket.PocketBuilder<?, ?> builder = generator.pocketBuilder(parameters)
				.virtualLocation(parameters.sourceVirtualLocation()); // TODO: virtualLocation thing still makes little sense
		generator.applyModifiers(parameters, builder);
		this.applyModifiers(parameters, builder);

		LazyPocketGenerator.currentlyGenerating = true;
		// ensure we aren't missing any chunks that were already loaded previously
		// for lazy gen
		Set<class_2791> alreadyLoadedChunks = StreamSupport.stream(parameters.world().method_14178().field_17254.method_17264().spliterator(), false).map(class_3193::method_16144).filter(Objects::nonNull).collect(Collectors.toSet());

		Pocket pocket = generator.prepareAndPlacePocket(parameters, builder);
		class_2338 originalOrigin = pocket.getOrigin();

		RiftManager manager = generator.getRiftManager(pocket);

		generator.applyModifiers(parameters, manager);

		this.applyModifiers(parameters, manager);

		if (pocket instanceof LazyGenerationPocket) {
			if (!(generator instanceof LazyPocketGenerator)) throw new RuntimeException("pocket was instance of LazyGenerationPocket but generator was not instance of LazyPocketGenerator");
			LazyGenerationPocket lazyPocket = (LazyGenerationPocket) pocket;
			LazyPocketGenerator clonedGenerator = ((LazyPocketGenerator) generator).cloneWithLazyModifiers(originalOrigin);
			if (setupLoot != null) clonedGenerator.setSetupLoot(setupLoot);

			attachLazyModifiers(clonedGenerator);
			clonedGenerator.attachToPocket(lazyPocket);
			lazyPocket.init();

			alreadyLoadedChunks.forEach(lazyPocket::chunkLoaded);

			LazyPocketGenerator.currentlyGenerating = false;

			while (!LazyPocketGenerator.generationQueue.isEmpty()) {
				class_2791 chunk = LazyPocketGenerator.generationQueue.remove();

				LazyCompatibleModifier.runQueuedModifications(chunk);
				MinecraftServer server = DimensionalDoorsInitializer.getServer();
				DimensionalDoorsInitializer.getServer().method_18858(new class_3738(server.method_3780(), () -> (lazyPocket).chunkLoaded(chunk)));
			}
			LazyCompatibleModifier.runLeftoverModifications(DimensionalDoorsInitializer.getWorld(lazyPocket.getWorld()));
		} else {
			LazyPocketGenerator.currentlyGenerating = false;
			LazyPocketGenerator.generationQueue.clear();
		}

		generator.setup(pocket, manager, parameters, setupLoot != null ? setupLoot : generator.isSetupLoot());

		return pocket;
	}

	@Override
	public PocketGeneratorReference peekNextPocketGeneratorReference(PocketGenerationContext parameters) {
		return this;
	}

	@Override
	public PocketGeneratorReference getNextPocketGeneratorReference(PocketGenerationContext parameters) {
		return this;
	}

	public abstract PocketGenerator peekReferencedPocketGenerator(PocketGenerationContext parameters);

	public abstract PocketGenerator getReferencedPocketGenerator(PocketGenerationContext parameters);

	@Override
	public abstract String toString();

	public void attachLazyModifiers(LazyPocketGenerator generator) {
		generator.attachLazyModifiers(modifierList.stream().filter(LazyModifier.class::isInstance).map(LazyModifier.class::cast).collect(Collectors.toList()));
	}
}
