package org.dimdev.dimdoors.world.decay;

import net.fabricmc.fabric.api.resource.SimpleSynchronousResourceReloadListener;
import net.minecraft.class_1301;
import net.minecraft.class_1937;
import net.minecraft.class_2248;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2487;
import net.minecraft.class_2520;
import net.minecraft.class_2680;
import net.minecraft.class_2960;
import net.minecraft.class_3218;
import net.minecraft.class_3300;
import net.minecraft.class_3419;
import net.minecraft.class_5321;
import net.minecraft.class_5819;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.dimdev.dimdoors.DimensionalDoorsInitializer;
import org.dimdev.dimdoors.api.util.ResourceUtil;
import org.dimdev.dimdoors.network.ExtendedServerPlayNetworkHandler;
import org.dimdev.dimdoors.network.packet.s2c.RenderBreakBlockS2CPacket;
import org.dimdev.dimdoors.sound.ModSoundEvents;

import java.util.*;
import java.util.concurrent.CompletableFuture;
import java.util.stream.Collectors;

/**
 * Provides methods for applying Limbo decay. Limbo decay refers to the effect that most blocks placed in Limbo
 * naturally change into stone, then cobble, then gravel, and finally Unraveled Fabric as time passes.
 */
public final class LimboDecay {
	private static final Logger LOGGER = LogManager.getLogger();
	private static final Map<class_5321<class_1937>, Set<DecayTask>> DECAY_QUEUE = new HashMap<>();

	private static final class_5819 RANDOM = class_5819.method_43047();

	/**
	 * Checks the blocks orthogonally around a given location (presumably the location of an Unraveled Fabric block)
	 * and applies Limbo decay to them. This gives the impression that decay spreads outward from Unraveled Fabric.
	 */
	public static void applySpreadDecay(class_3218 world, class_2338 pos) {
		//Check if we randomly apply decay spread or not. This can be used to moderate the frequency of
		//full spread decay checks, which can also shift its performance impact on the game.
		if (RANDOM.method_43058() < DimensionalDoorsInitializer.getConfig().getLimboConfig().decaySpreadChance) {
			class_2680 origin = world.method_8320(pos);

			//Apply decay to the blocks above, below, and on all four sides.
			// TODO: make max amount configurable
			int decayAmount = RANDOM.method_43048(5) + 1;
			List<class_2350> directions = new ArrayList<>(Arrays.asList(class_2350.values()));
			for (int i = 0; i < decayAmount; i++) {
				decayBlock(world, pos.method_10093(directions.remove(RANDOM.method_43048(5 - i))), origin);
			}
		}
	}

	/**
	 * Checks if a block can be decayed and, if so, changes it to the next block ID along the decay sequence.
	 */
	private static void decayBlock(class_3218 world, class_2338 pos, class_2680 origin) {
		class_2680 target = world.method_8320(pos);

		Collection<DecayPattern> patterns = DecayLoader.getInstance().getPatterns(target.method_26204());

		if(patterns == null || patterns.isEmpty()) {
			return;
		}


		for(DecayPattern pattern : patterns) {
			if (!pattern.test(world, pos, origin, target)) {
				continue;
			}
			world.method_18766(class_1301.method_5909(pos.method_10263(), pos.method_10264(), pos.method_10260(), 100)).forEach(player -> {
				ExtendedServerPlayNetworkHandler.get(player.field_13987).getDimDoorsPacketHandler().sendPacket(new RenderBreakBlockS2CPacket(pos, 5));
			});
			world.method_43128(null, pos.method_10263(), pos.method_10264(), pos.method_10260(), ModSoundEvents.TEARING, class_3419.field_15245, 0.5f, 1f);
			queueDecay(world, pos, origin, pattern, DimensionalDoorsInitializer.getConfig().getLimboConfig().limboDecay);
			break;
		}
	}

	public static void queueDecay(class_3218 world, class_2338 pos, class_2680 origin, DecayPattern pattern, int delay) {
		DecayTask task = new DecayTask(pos, origin, pattern, delay);
		if (delay <= 0) {
			task.process(world);
		} else {
			DECAY_QUEUE.computeIfAbsent(world.method_27983(), k -> new HashSet<>()).add(task);
		}
	}

	public static void tick(class_3218 world) {
		class_5321<class_1937> key = world.method_27983();
		if (DECAY_QUEUE.containsKey(key)) {
			Set<DecayTask> tasks = DECAY_QUEUE.get(key);
			Set<DecayTask> tasksToRun = tasks.stream().filter(DecayTask::reduceDelayIsDone).collect(Collectors.toSet());
			tasks.removeAll(tasksToRun);
			tasksToRun.forEach(task -> task.process(world));
		}
	}

	public static class DecayLoader implements SimpleSynchronousResourceReloadListener {
		private static final Logger LOGGER = LogManager.getLogger();
		private static final DecayLoader INSTANCE = new DecayLoader();
		private final Map<class_2248, List<DecayPattern>> patterns = new HashMap();

		private DecayLoader() {
		}

		public static DecayLoader getInstance() {
			return INSTANCE;
		}

		@Override
		public void method_14491(class_3300 manager) {
			patterns.clear();
			CompletableFuture<List<DecayPattern>> futurePatternList = ResourceUtil.loadResourcePathToCollection(manager, "decay_patterns", ".json", new ArrayList<>(), ResourceUtil.NBT_READER.andThenReader(this::loadPattern));
			for (DecayPattern pattern : futurePatternList.join()) {
				for (class_2248 block : pattern.constructApplicableBlocks()) {
					patterns.computeIfAbsent(block, (b) -> new ArrayList<>());
					patterns.get(block).add(pattern);
				}
			}
		}

		private DecayPattern loadPattern(class_2520 nbt, class_2960 ignored) {
			return DecayPattern.deserialize((class_2487) nbt);
		}

		public Collection<DecayPattern> getPatterns(class_2248 block) {
			return patterns.get(block);
		}

		@Override
		public class_2960 getFabricId() {
			return new class_2960("dimdoors", "decay_pattern");
		}
	}

	private static class DecayTask {
		private final class_2338 pos;
		private final class_2680 origin;
		private final DecayPattern processor;
		private int delay;


		public DecayTask(class_2338 pos, class_2680 origin, DecayPattern processor, int delay) {
			this.pos = pos;
			this.origin = origin;
			this.processor = processor;
			this.delay = delay;
		}

		public boolean reduceDelayIsDone() {
			return --delay <= 0;
		}

		public void process(class_3218 world) {
			class_2680 target = world.method_8320(pos);
			if (world.method_22340(pos) && processor.test(world, pos, origin, target)) {
				world.method_18766(class_1301.method_5909(pos.method_10263(), pos.method_10264(), pos.method_10260(), 100)).forEach(player -> {
					ExtendedServerPlayNetworkHandler.get(player.field_13987).getDimDoorsPacketHandler().sendPacket(new RenderBreakBlockS2CPacket(pos, -1));
				});
				world.method_43128(null, pos.method_10263(), pos.method_10264(), pos.method_10260(), target.method_26231().method_10595(), class_3419.field_15245, 0.5f, 1f);
				processor.process(world, pos, origin, world.method_8320(pos));
			}
		}
	}
}
