package org.dimdev.dimdoors.world.pocket;

import java.util.HashMap;
import java.util.Map;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.concurrent.CompletableFuture;
import java.util.stream.Collectors;
import net.minecraft.class_1937;
import net.minecraft.class_2338;
import net.minecraft.class_2378;
import net.minecraft.class_2382;
import net.minecraft.class_2487;
import net.minecraft.class_2960;
import net.minecraft.class_3545;
import net.minecraft.class_5321;
import org.dimdev.dimdoors.DimensionalDoorsInitializer;
import org.dimdev.dimdoors.api.util.math.GridUtil;
import org.dimdev.dimdoors.world.pocket.type.AbstractPocket;
import org.dimdev.dimdoors.world.pocket.type.IdReferencePocket;
import org.dimdev.dimdoors.world.pocket.type.Pocket;
import org.jetbrains.annotations.TestOnly;

public class PocketDirectory {
	int gridSize; // Determines how much pockets in their dimension are spaced
	int privatePocketSize;
	int publicPocketSize;
	Map<Integer, AbstractPocket<?>> pockets;
	private SortedMap<Integer, Integer> nextIDMap;
	class_5321<class_1937> worldKey;

	public PocketDirectory(class_5321<class_1937> worldKey) {
		this.gridSize = DimensionalDoorsInitializer.getConfig().getPocketsConfig().pocketGridSize;
		this.worldKey = worldKey;
		this.nextIDMap = new TreeMap<>();
		this.pockets = new HashMap<>();
	}

	@TestOnly
	public PocketDirectory(class_5321<class_1937> worldKey, int gridSize) {
		this.gridSize = gridSize;
		this.worldKey = worldKey;
		this.nextIDMap = new TreeMap<>();
		this.pockets = new HashMap<>();
	}

	public static PocketDirectory readFromNbt(String id, class_2487 nbt) {
		PocketDirectory directory = new PocketDirectory(class_5321.method_29179(class_2378.field_25298, new class_2960(id)));
		// no need to parallelize
		directory.gridSize = nbt.method_10550("grid_size");
		directory.privatePocketSize = nbt.method_10550("private_pocket_size");
		directory.publicPocketSize = nbt.method_10550("public_pocket_size");
		// same thing, too short anyways
		class_2487 nextIdMapNbt = nbt.method_10562("next_id_map");
		directory.nextIDMap.putAll(nextIdMapNbt.method_10541().stream().collect(Collectors.toMap(Integer::parseInt, nextIdMapNbt::method_10550)));

		class_2487 pocketsNbt = nbt.method_10562("pockets");
		directory.pockets = pocketsNbt.method_10541().stream().unordered().map(key -> {
			class_2487 pocketNbt = pocketsNbt.method_10562(key);
			return CompletableFuture.supplyAsync(() -> new class_3545<>(Integer.parseInt(key), AbstractPocket.deserialize(pocketNbt)));
		}).parallel().map(CompletableFuture::join).collect(Collectors.toConcurrentMap(class_3545::method_15442, class_3545::method_15441));

		return directory;
	}

	public class_2487 writeToNbt() {
		class_2487 nbt = new class_2487();
		nbt.method_10569("grid_size", this.gridSize);
		nbt.method_10569("private_pocket_size", this.privatePocketSize);
		nbt.method_10569("public_pocket_size", this.publicPocketSize);

		class_2487 nextIdMapNbt = new class_2487();
		this.nextIDMap.forEach((key, value) -> nextIdMapNbt.method_10569(key.toString(), value));
		nbt.method_10566("next_id_map", nextIdMapNbt);

		class_2487 pocketsNbt = new class_2487();
		this.pockets.entrySet().parallelStream().unordered().map(entry -> CompletableFuture.supplyAsync(() -> new class_3545<>(entry.getKey().toString(), entry.getValue().toNbt(new class_2487()))))
				.map(CompletableFuture::join).sequential().forEach(pair -> pocketsNbt.method_10566(pair.method_15442(), pair.method_15441()));
		nbt.method_10566("pockets", pocketsNbt);

		return nbt;
	}

	/**
	 * Create a new blank pocket.
	 *
	 * @return The newly created pockets
	 */
	public <T extends Pocket> T newPocket(Pocket.PocketBuilder<?, T> builder) {
		class_2382 size = builder.getExpectedSize();
		int longest = Math.max(Math.max(size.method_10263(), size.method_10260()), 1);
		longest = (Math.floorDiv(longest - 1, gridSize * 16)) + 1;

		int base3Size = 1;
		while (longest > base3Size) {
			base3Size *= 3;
		}

		int squaredSize = base3Size * base3Size;

		int cursor = nextIDMap.headMap(base3Size+1).values().stream().mapToInt(num -> num).max().orElse(0);
		cursor = cursor - Math.floorMod(cursor, squaredSize);

		Pocket pocketAt = getPocket(cursor);
		while (pocketAt != null) {
			size = pocketAt.getSize();
			longest = Math.max(size.method_10263(), size.method_10260());
			longest = (longest / (gridSize * 16)) + 1;

			int pocketBase3Size = 1;
			while (longest > pocketBase3Size) {
				pocketBase3Size *= 3;
			}

			cursor += Math.max(squaredSize, pocketBase3Size * pocketBase3Size);
			pocketAt = getPocket(cursor);
		}

		cursor = cursor + squaredSize - 1; // we actually want to use the last id of
		// the assigned grid space since it is in the bottom left corner

		T pocket = builder
				.id(cursor)
				.world(worldKey)
				.range(squaredSize)
				.offsetOrigin(idToCenteredPos(cursor, base3Size, builder.getExpectedSize()))
				.build();

		nextIDMap.put(base3Size, cursor + squaredSize);
		addPocket(pocket);

		IdReferencePocket.IdReferencePocketBuilder idReferenceBuilder = IdReferencePocket.builder();
		for (int i = 1; i < squaredSize; i++) {
			addPocket(idReferenceBuilder
					.id(cursor - i)
					.world(worldKey)
					.referencedId(cursor)
					.build());
		}
		return pocket;
	}

	private void addPocket(AbstractPocket<?> pocket) {
		pockets.put(pocket.getId(), pocket);
	}

	// TODO: rework this method to remove references as well
	public void removePocket(int id) {
		this.pockets.remove(id);
	}

	/**
	 * Gets the pocket that occupies the GridPos which a certain ID represents, or null if there is no pocket at that GridPos.
	 *
	 * @return The pocket which occupies the GridPos represented by that ID, or null if there was no pocket occupying that GridPos.
	 */
	public Pocket getPocket(int id) {
		AbstractPocket<?> pocket = this.pockets.get(id);
		return pocket == null ? null : pocket.getReferencedPocket(this);
	}

	public <P extends Pocket> P getPocket(int id, Class<P> clazz) {
		Pocket pocket = getPocket(id);
		if (clazz.isInstance(pocket)) return clazz.cast(pocket);
		return null;
	}

	public GridUtil.GridPos idToGridPos(int id) {
		return GridUtil.idToGridPos(id);
	}

	public int gridPosToID(GridUtil.GridPos pos) {
		return GridUtil.gridPosToID(pos);
	}

	/**
	 * Calculates the default BlockPos where a pocket should be based on the ID. Use this only for placing
	 * pockets, and use Pocket.getGridPos() for getting the position
	 *
	 * @param id The ID of the pocket
	 * @return The BlockPos of the pocket
	 */
	public class_2338 idToPos(int id) {
		GridUtil.GridPos pos = this.idToGridPos(id);
		return new class_2338(pos.x * this.gridSize * 16, 0, pos.z * this.gridSize * 16);
	}

	public class_2338 idToCenteredPos(int id, int base3Size, class_2382 expectedSize) {
		GridUtil.GridPos pos = this.idToGridPos(id);
		// you actually need the "/ 2 * 16" here. "*8" would not work the same since it doesn't guarantee chunk alignment
		return new class_2338((pos.x * this.gridSize * 16) + (base3Size * this.gridSize - expectedSize.method_10263() / 16) / 2 * 16, 0, (pos.z * this.gridSize * 16) + (base3Size * this.gridSize - expectedSize.method_10260() / 16) / 2 * 16);
	}

	/**
	 * Calculates the ID of a pocket at a certain BlockPos.
	 *
	 * @param pos The position
	 * @return The ID of the pocket, or -1 if there is no pocket at that location
	 */
	public int posToID(class_2338 pos) {
		return this.gridPosToID(new GridUtil.GridPos(Math.floorDiv(pos.method_10263(), this.gridSize * 16), Math.floorDiv(pos.method_10260(), this.gridSize * 16)));
	}

	public Pocket getPocketAt(class_2338 pos) { // TODO: use BlockPos
		return this.getPocket(this.posToID(pos));
	}

	public boolean isWithinPocketBounds(class_2338 pos) {
		Pocket pocket = this.getPocketAt(pos);
		return pocket != null && pocket.isInBounds(pos);
	}

	public int getGridSize() {
		return this.gridSize;
	}

	public int getPrivatePocketSize() {
		return this.privatePocketSize;
	}

	public int getPublicPocketSize() {
		return this.publicPocketSize;
	}

	public Map<Integer, AbstractPocket<?>> getPockets() {
		return this.pockets;
	}
}


